/*
BB84-full-prac.h
Written by Walter O. Krawec and Sam A. Markelon
Copyright (c) 2018

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "../GateEvolve.h"

// six state protocol
namespace QKD
{
namespace GateGA
{

namespace Protocols
{
  class BB84Fullprac : public QKD::GateGA::Protocol
  {
  public:
    virtual ~BB84Fullprac()
    {
    }

    virtual bool setup()
    {
    }

    virtual int numChannels()
    {
      return 1;
    }

    virtual void printStats(std::vector < std::list<Gate> >& gates)
    {
      double rate = computeKeyRate(gates);
      std::vector <double> stats;
      computeErrorRate(gates, stats);
      double avgNoise = 0;
      std::cout << "\t\tBB84:rate=" << rate << ":[";
      for(int i=0; i<stats.size(); ++i){
	avgNoise += stats[i];
	std::cout << stats[i] << ", ";
      }
      std::cout << "avg=" << avgNoise/(double)stats.size();
      std::cout << "]\n";
    }

    virtual double computeKeyRate(std::vector < std::list<Gate> >& gates)
    {
      const int NUM_WIRES=5+numAttackWires(); // A, B, Pub1, Pub2, T, E_guess E_aux1

      const int A = 0;
      const int B = 1;
      const int P1 = 2;
      const int P2 = 3;
      const int T = 4;
      const int E = 5;
      const int Eaux = 6;
      // simulate the channel: A sends |0> or |1> or |+> or |-> or |0Y>, |1Y>:

      quantum::DensityList rho;

      setupWires(rho, NUM_WIRES); 

      quantum::KetBra kb(NUM_WIRES);
      kb.p = 1.0/6.0;
      rho.density.push_back(kb);

      kb.ket[A] = 1;
      kb.bra[A] = 1;
      kb.ket[T] = 1;
      kb.bra[T] = 1;
      rho.density.push_back(kb);

      kb.ket[A] = 0;
      kb.bra[A] = 0;
      kb.ket[P1] = 1;
      kb.bra[P1] = 1;
      kb.ket[T] = 0;
      kb.bra[T] = 0;
      rho.density.push_back(kb);


      kb.ket[A] = 1;
      kb.bra[A] = 1;
      kb.ket[P1] = 1;
      kb.bra[P1] = 1;
      kb.ket[T] = 1;
      kb.bra[T] = 1;
      rho.density.push_back(kb);


      kb.ket[A] = 0;
      kb.bra[A] = 0;
      kb.ket[P1] = 0;
      kb.bra[P1] = 0;
      kb.ket[P2] = 1;
      kb.bra[P2] = 1;
      kb.ket[T] = 0;
      kb.bra[T] = 0;
      rho.density.push_back(kb);

      kb.ket[A] = 1;
      kb.bra[A] = 1;
      kb.ket[P1] = 0;
      kb.bra[P1] = 0;
      kb.ket[P2] = 1;
      kb.bra[P2] = 1;
      kb.ket[T] = 1;
      kb.bra[T] = 1;
      rho.density.push_back(kb);



      algebra::mat H;
      H.create(2);
      H(0,0) = 1.0/sqrt(2.0);
      H(0,1) = 1.0/sqrt(2.0);
      H(1,0) = 1.0/sqrt(2.0);
      H(1,1) = -1.0/sqrt(2.0);

      algebra::mat Y;
      Y.create(2);
      Y(0,0) = 1.0/sqrt(2.0);
      Y(0,1) = 1.0/sqrt(2.0);
      Y(1,0) = 1.0/sqrt(2.0)*I;
      Y(1,1) = -1.0/sqrt(2.0)*I;

      algebra::mat Yinv;
      Y.transpose(Yinv);

      rho.applyConditionalOp(P1, 1, T, H);
      rho.applyConditionalOp(P2, 1, T, Y);


      // Eve attacks:

      applyAttack(gates[0], rho, T);

      // Eve measures:
      for(int i=E; i<NUM_WIRES; ++i)
	rho.measure(i);

      for(int i=Eaux; i<NUM_WIRES; ++i)
	rho.trace(i);

      // Bob measures:
      rho.applyConditionalOp(P1, 1, T, H);
      rho.applyConditionalOp(P2, 1, T, Yinv);
      rho.measureAndSave(T,B);
      rho.trace(T);

      double ab00 = rho.calculatePr(A, 0, B, 0);
      double ab01 = rho.calculatePr(A, 0, B, 1);
      double ab10 = rho.calculatePr(A, 1, B, 0);
      double ab11 = rho.calculatePr(A, 1, B, 1);


      //rho.print(std::cout);
      //std::cout << "\n\n";

      rho.trace(B);


      // E aux not used (opt. measure these instead)
      //for(int i=5; i<NUM_WIRES; ++i)
      //rho.trace(i);
      //rho.print(std::cout);


      algebra::mat M;
      rho.computeDensityMatFast(M);
      double HAE = M.entropy();
      rho.trace(0);
      rho.computeDensityMatFast(M);
      double HE = M.entropy();
      HAE = HAE - HE;

      double HAB = entropy(ab00, ab01, ab10, ab11) - entropy(ab00+ab10);

      return HAE - HAB;
    }

    virtual void computeErrorRate(std::vector < std::list<Gate> >& gates, std::vector <double>& errorStats)
    {
      errorStats.resize(1);
      const int NUM_WIRES=5+numAttackWires(); // A, B, Pub1, Pub2, T, E_guess E_aux1

      const int A = 0;
      const int B = 1;
      const int P1 = 2;
      const int P2 = 3;
      const int T = 4;
      const int E = 5;
      const int Eaux = 6;
      // simulate the channel: A sends |0> or |1> or |+> or |-> or |0Y>, |1Y>:

      quantum::DensityList rho;

      setupWires(rho, NUM_WIRES); 

      quantum::KetBra kb(NUM_WIRES);
      kb.p = 1.0/6.0;
      rho.density.push_back(kb);

      kb.ket[A] = 1;
      kb.bra[A] = 1;
      kb.ket[T] = 1;
      kb.bra[T] = 1;
      rho.density.push_back(kb);

      kb.ket[A] = 0;
      kb.bra[A] = 0;
      kb.ket[P1] = 1;
      kb.bra[P1] = 1;
      kb.ket[T] = 0;
      kb.bra[T] = 0;
      rho.density.push_back(kb);


      kb.ket[A] = 1;
      kb.bra[A] = 1;
      kb.ket[P1] = 1;
      kb.bra[P1] = 1;
      kb.ket[T] = 1;
      kb.bra[T] = 1;
      rho.density.push_back(kb);


      kb.ket[A] = 0;
      kb.bra[A] = 0;
      kb.ket[P1] = 0;
      kb.bra[P1] = 0;
      kb.ket[P2] = 1;
      kb.bra[P2] = 1;
      kb.ket[T] = 0;
      kb.bra[T] = 0;
      rho.density.push_back(kb);

      kb.ket[A] = 1;
      kb.bra[A] = 1;
      kb.ket[P1] = 0;
      kb.bra[P1] = 0;
      kb.ket[P2] = 1;
      kb.bra[P2] = 1;
      kb.ket[T] = 1;
      kb.bra[T] = 1;
      rho.density.push_back(kb);



      algebra::mat H;
      H.create(2);
      H(0,0) = 1.0/sqrt(2.0);
      H(0,1) = 1.0/sqrt(2.0);
      H(1,0) = 1.0/sqrt(2.0);
      H(1,1) = -1.0/sqrt(2.0);

      algebra::mat Y;
      Y.create(2);
      Y(0,0) = 1.0/sqrt(2.0);
      Y(0,1) = 1.0/sqrt(2.0);
      Y(1,0) = 1.0/sqrt(2.0)*I;
      Y(1,1) = -1.0/sqrt(2.0)*I;

      algebra::mat Yinv;
      Y.transpose(Yinv);

      rho.applyConditionalOp(P1, 1, T, H);
      rho.applyConditionalOp(P2, 1, T, Y);


      // Eve attacks:

      applyAttack(gates[0], rho, T);

      // Eve measures:
      for(int i=E; i<NUM_WIRES; ++i)
	rho.measure(i);

      for(int i=Eaux; i<NUM_WIRES; ++i)
	rho.trace(i);

      // Bob measures:
      rho.applyConditionalOp(P1, 1, T, H);
      rho.applyConditionalOp(P2, 1, T, Yinv);
      rho.measureAndSave(T,B);
      rho.trace(T);

      double ab00 = rho.calculatePr(A, 0, B, 0);
      double ab01 = rho.calculatePr(A, 0, B, 1);
      double ab10 = rho.calculatePr(A, 1, B, 0);
      double ab11 = rho.calculatePr(A, 1, B, 1);


      errorStats[0] = ab01+ab10;
    }


    virtual void computeErrorRate_old(std::vector < std::list<Gate> >& gates, std::vector <double>& errorStats)
    {
      errorStats.resize(3);  // Z and X and Y basis noise
      const int NUM_WIRES=3+numAttackWires(); // A, B, T, E_guess E_aux1

      /****** Z *******/
      {
	// simulate the channel: A sends |0> or |1>:
	quantum::DensityList rho;

	setupWires(rho, NUM_WIRES); 

	quantum::KetBra kb(NUM_WIRES);
	kb.p = .5;
	
	rho.density.push_back(kb);
	kb.ket[0] = 1;
	kb.bra[0] = 1;
	kb.ket[2] = 1;
	kb.bra[2] = 1;
	
	rho.density.push_back(kb);
	
	// Eve attacks:
	
	applyAttack(gates[0], rho, 2);

	// Eve measures:
	for(int i=3; i<NUM_WIRES; ++i)
	  rho.measure(i);
	
	// Bob measures:
	rho.measureAndSave(2,1);
	
	double p01Z = rho.calculatePr(0, 0, 1, 1);
	double p10Z = rho.calculatePr(0, 1, 1, 0);

	errorStats[0] = p01Z+p10Z;
      }

      /****** X *******/
      {
	// simulate the channel: A sends |0> or |1>:
	quantum::DensityList rho;

	setupWires(rho, NUM_WIRES); 

	quantum::KetBra kb(NUM_WIRES);
	kb.p = .5;
	
	rho.density.push_back(kb);
	kb.ket[0] = 1;
	kb.bra[0] = 1;
	kb.ket[2] = 1;
	kb.bra[2] = 1;
	
	rho.density.push_back(kb);

	rho.applyHadamard(2);
	
	// Eve attacks:
	
	applyAttack(gates[0], rho, 2);

	// Eve measures:
	for(int i=3; i<NUM_WIRES; ++i)
	  rho.measure(i);
	
	// Bob measures:
	rho.applyHadamard(2);
	rho.measureAndSave(2,1);
	
	double p01X = rho.calculatePr(0, 0, 1, 1);
	double p10X = rho.calculatePr(0, 1, 1, 0);

	errorStats[1] = p01X+p10X;
      }

      /****** X *******/
      {
	// simulate the channel: A sends |0> or |1>:
	algebra::mat Y;
	Y.create(2);
	Y(0,0) = 1.0/sqrt(2.0);
	Y(0,1) = 1.0/sqrt(2.0);
	Y(1,0) = 1.0/sqrt(2.0)*I;
	Y(1,1) = -1.0/sqrt(2.0)*I;
	
	algebra::mat Yinv;
	Y.transpose(Yinv);

	quantum::DensityList rho;

	setupWires(rho, NUM_WIRES); 

	quantum::KetBra kb(NUM_WIRES);
	kb.p = .5;
	
	rho.density.push_back(kb);
	kb.ket[0] = 1;
	kb.bra[0] = 1;
	kb.ket[2] = 1;
	kb.bra[2] = 1;
	
	rho.density.push_back(kb);

	rho.applyOp(2, Y);
	
	// Eve attacks:
	
	applyAttack(gates[0], rho, 2);

	// Eve measures:
	for(int i=3; i<NUM_WIRES; ++i)
	  rho.measure(i);
	
	// Bob measures:
	rho.applyOp(2, Yinv);
	rho.measureAndSave(2,1);
	
	double p01Y = rho.calculatePr(0, 0, 1, 1);
	double p10Y = rho.calculatePr(0, 1, 1, 0);

	errorStats[2] = p01Y+p10Y;
      }
    }
  };
}

}
}
